from datetime import datetime
from pathlib import Path
import sys
sys.path.append(".")
import unittest

from frads import geom
import pyradiance as pr
from frads import sky
from frads import utils
from frads.sky import WeaData
from frads.sky import WeaMetaData
import logging

logger = logging.getLogger()
logger.setLevel(50)


test_dir = Path(__file__).resolve().parent
epw_path = test_dir / "Resources" / "USA_CA_Oakland.Intl.AP.724930_TMY3.epw"
wea_path = test_dir / "Resources" / "oak.wea"

class TestSky(unittest.TestCase):

    # def test_basis_glow():
    #     basis = "r1"
    #     result = sky.basis_glow(basis)
    #     assert result == ("#@rfluxmtx h=u\n\n"
    #                       "void glow groundglow\n"
    #                       "0\n0\n4 1 1 1 0\n\n"
    #                       "groundglow source ground\n"
    #                       "0\n0\n4 0 0 -1 180\n\n"
    #                       "#@rfluxmtx u=+Y h=r1\n\n"
    #                       "void glow skyglow\n"
    #                       "0\n0\n4 1 1 1 0\n\n"
    #                       "skyglow source sky\n"
    #                       "0\n0\n4 0 0 1 180\n")


    # def test_gen_sun_source_full():
    #     """Generate sun sources for matrix generation."""
    #     reinsrc4_path = test_dir / "Resources" / "reinsrc4.rad"
    #     mf = 4
    #     prim_str, mod_str = sky.gen_sun_source_full(mf)
    #     prims = pr.parse_primitive(prim_str)
    #     answer_prims = utils.unpack_primitives(reinsrc4_path)
    #     for aprim, prim in zip(prims, answer_prims):
    #         assert str(aprim) == str(prim)

    def test_gendaymtx(self):
        """Generate a psuedo reinhart 6 sun matrix file given lat, lon, etc..."""
        # sky.gendaymtx(
        #     sun_mtx, 6, data=wea_data, meta=wea_metadata, direct=True, onesun=True
        # )
        pass

    def test_filter_wea(self):
        """."""
        # wea_data, _ = sky.filter_wea(
        #     wea_data, wea_metadata, start_hour=6, end_hour=20,
        #     remove_zero=True, daylight_hours_only=True)
        pass

    # def test_check_sun_above_horizon():
    #     sky.check_sun_above_horizon()
    #     pass


    def test_solar_angle(self):
        # sky.filter_data_by_direct_sun()
        pass

    def test_start_end_hour(self):
        """Remove wea data entries outside of the
        start and end hour."""
        # sh = None
        # eh = None
        # data = None
        # result = makesky.start_end_hour(sh, eh, data)
        pass

    def test_check_sun_above_horizon(self):
        """Remove non-daylight hour entries."""
        pass

    def test_filter_wea_zero_entry(self):
        """Remove wea data entries with zero solar luminance.
        If window normal supplied, eliminate entries not seen by window.
        Solar luminance determined using Perez sky model.
        Window field of view is 176 deg with 2 deg tolerance on each side.
        """
        pass

    def test_gen_sun_source_culled(self):
        # mf = 6
        # suns, mod, full_mod = sky.gen_sun_source_culled(mf, smx_path=self.sunmtx_path)
        # assert Equal(len(suns.splitlines()), 5186)
        # assert Equal(len(mod.splitlines()), 615)
        # sun_mtx.unlink()
        pass

    # def test_filter_data_by_direct_sun():
    #     meta = WeaMetaData("Oakland", "USA", 37.72, 122.22, 120, 2)
    #     data = [
    #         WeaData(datetime(2000, 1, 1, 6, 30), 0, 0),
    #         WeaData(datetime(2000, 1, 1, 7, 30), 71, 1),
    #         WeaData(datetime(2000, 1, 1, 8, 30), 463,  41),
    #         WeaData(datetime(2000, 1, 1, 13, 30), 717, 110),
    #         WeaData(datetime(2000, 7, 1, 15, 30), 473,  95),
    #         WeaData(datetime(2000, 7, 1, 16, 30), 244,  16),
    #         WeaData(datetime(2000, 1, 1, 17, 30),   0,   0),
    #     ]
    #     wnormal = geom.Vector(0, 1, 0)
    #     filtered = sky.filter_data_by_direct_sun(data, meta, window_normal=[wnormal])
    #     answer = [
    #         WeaData(datetime(2000, 1, 1,  7, 30),  71,   1),
    #         WeaData(datetime(2000, 1, 1,  8, 30), 463,  41),
    #         WeaData(datetime(2000, 1, 1, 13, 30), 717, 110),
    #         WeaData(datetime(2000, 7, 1, 15, 30), 473,  95),
    #     ]
    #     for res, ans in zip(filtered, answer):
    #         assert res == ans

    def test_parser_epw(self):
        with open(epw_path) as rdr:
            wea_metadata, wea_data = sky.parse_epw(rdr.read())
        self.assertEqual( wea_data[0].time.month, 1)
        self.assertEqual( wea_data[-1].time.month, 12)
        self.assertEqual( wea_data[-1].time.day, 31)
        self.assertEqual( wea_metadata.latitude, 37.72)

    def test_parser_wea(self):
        with open(wea_path) as rdr:
            wea_metadata, wea_data = sky.parse_wea(rdr.read())
        self.assertEqual(wea_data[0].time.month, 1)
        self.assertEqual(wea_data[-1].time.month, 12)
        self.assertEqual(wea_data[-1].time.day, 31)
        self.assertEqual(wea_data[-1].dni, 0)
        self.assertEqual(wea_metadata.latitude, 37.72)

if __name__ == "__main__":
    unittest.main()



from datetime import datetime
from pathlib import Path
import unittest

import pyradiance as pr
from pyenergyplus.dataset import ref_models, weather_files
import frads as fr


class TestMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.resources_dir = Path(__file__).parent / "Resources"
        cls.objects_dir = Path(__file__).parent / "Objects"

    def setUp(self):
        self.cfg = {
            "settings": {
                "method": "2phase",
                "sky_basis": "r1",
                "epw_file": "",
                "wea_file": self.resources_dir / "oak.wea",
                "sensor_sky_matrix": ["-ab", "0"],
                "view_sky_matrix": ["-ab", "0"],
                "sensor_window_matrix": ["-ab", "0"],
                "view_window_matrix": ["-ab", "0"],
                "daylight_matrix": ["-ab", "0"],
            },
            "model": {
                "scene": {
                    "files": [
                        self.objects_dir / "walls.rad",
                        self.objects_dir / "ceiling.rad",
                        self.objects_dir / "floor.rad",
                        self.objects_dir / "ground.rad",
                    ]
                },
                "windows": {
                    "upper_glass": {
                        "file": self.objects_dir / "upper_glass.rad",
                        "matrix_name": "blinds30",
                    },
                    "lower_glass": {
                        "file": self.objects_dir / "lower_glass.rad",
                        "matrix_name": "blinds30",
                    },
                },
                "materials": {
                    "files": [self.objects_dir / "materials.mat"],
                    "matrices": {
                        "blinds30": {"matrix_file": self.resources_dir / "blinds30.xml"}
                    },
                },
                "sensors": {
                    "wpi": {"file": self.resources_dir / "grid.txt"},
                    "view1": {
                        "data": [[17, 5, 4, 1, 0, 0]],
                    },
                },
                "views": {
                    "view1": {
                        "file": self.resources_dir / "v1a.vf",
                        "xres": 16,
                        "yres": 16,
                    }
                },
                "surfaces": {},
            },
        }

        self.scene = fr.SceneConfig(
                files=[
                    self.objects_dir / "walls.rad",
                    self.objects_dir / "ceiling.rad",
                    self.objects_dir / "floor.rad",
                    self.objects_dir / "ground.rad",
                ]
            )


        self.window_1 = fr.WindowConfig(
                file=self.objects_dir / "upper_glass.rad",
                matrix_name="blinds30",
            )


        self.window_2 = fr.WindowConfig(
                file=self.objects_dir / "upper_glass.rad",
                # matrix_name="blinds30",
            )


        self.materials = fr.MaterialConfig(
                files=[self.objects_dir / "materials.mat"],
                matrices={"blinds30": {"matrix_file": self.resources_dir / "blinds30.xml"}},
            )


        self.wpi = fr.SensorConfig(file=self.resources_dir / "grid.txt")


        self.sensor_view_1 = fr.SensorConfig( data=[[17, 5, 4, 1, 0, 0]],)


        self.view_1 = fr.ViewConfig( file=self.resources_dir / "v1a.vf", xres=16, yres=16,)


    def test_model1(self):
        model = fr.Model(
            scene=self.scene,
            windows={"window_1": self.window_1},
            materials=self.materials,
            sensors={"wpi": self.wpi, "view1": self.sensor_view_1},
            views={"view_1": self.view_1},
        )
        self.assertEqual(model.scene.files, self.scene.files)
        self.assertEqual(model.windows["window_1"].file, self.window_1.file)
        self.assertEqual(model.windows["window_1"].matrix_name, self.window_1.matrix_name)
        self.assertEqual(model.materials.files, self.materials.files)
        self.assertIn(model.windows["window_1"].matrix_name, model.materials.matrices)
        self.assertEqual(model.sensors["wpi"].file, self.wpi.file)
        self.assertEqual(model.sensors["view_1"].data[0][0], self.sensor_view_1.data[0][0])
        self.assertEqual(model.views["view_1"].file, self.view_1.file)
        self.assertEqual(model.views["view_1"].xres, self.view_1.xres)
        self.assertEqual(model.views["view_1"].yres, self.view_1.yres)


    def test_model2(self):
        # auto-generate view_1 in sensors from view_1 in views
        model = fr.Model(
            materials=self.materials,
            sensors={"wpi": self.wpi},
            views={"view_1": self.view_1},
        )
        self.assertIn("view_1", model.sensors)
        self.assertEqual(model.sensors["view_1"].data[0][0], model.views["view_1"].view.vp[0])
        self.assertTrue(isinstance(model.scene, fr.SceneConfig))
        self.assertTrue(isinstance(model.windows, dict))
        self.assertEqual(model.scene.files, [])
        self.assertEqual(model.scene.bytes, b"")
        self.assertEqual(model.windows, {})


    def test_model3(self):
        # same name view and sensor but different position and direction
        sensor_view_2 = fr.SensorConfig(
            data=[[1, 5, 4, 1, 0, 0]],
        )

        with self.assertRaises(ValueError):
            fr.Model(
                scene=self.scene,
                windows={"window_1": self.window_1},
                materials=self.materials,
                sensors={"wpi": self.wpi, "view_1": sensor_view_2},
                views={"view_1": self.view_1},
            )


    def test_model4(self):
        # window matrix name not in materials
        materials = fr.MaterialConfig(files=[self.objects_dir / "materials.mat"])
        with self.assertRaises(ValueError):
            fr.Model(
                scene=self.scene,
                windows={"window_1": self.window_1},
                materials=materials,
                sensors={"wpi": self.wpi, "view_1": self.sensor_view_1},
                views={"view_1": self.view_1},
            )


    def test_no_sensors_views_surfaces_specified(self):
        settings = fr.Settings()
        model = fr.Model(
            scene=self.scene,
            windows={"window_1": self.window_1},
            materials=self.materials,
        )
        with self.assertRaises(ValueError):
            fr.WorkflowConfig(settings, model)


    def test_windows_not_specified_for_3phase_or_5phase_method(self):
        settings = fr.Settings()
        model = fr.Model(
            scene=self.scene,
            # windows={"window_1": window_1},
            materials=self.materials,
            sensors={"wpi": self.wpi},
            views={"view_1": self.view_1},
        )
        with self.assertRaises(ValueError):
            fr.WorkflowConfig(settings, model)


    def test_three_phase2(self):
        model = fr.Model(
            scene=self.scene,
            windows={"window_1": self.window_2},  # window_2 has no matrix_name
            materials=self.materials,
            sensors={"wpi": self.wpi, "view_1": self.sensor_view_1},
            views={"view_1": self.view_1},
        )
        settings = fr.Settings()
        settings.sensor_window_matrix = ['-ab', '1']
        settings.daylight_matrix = ['-ab', '1']

        cfg = fr.WorkflowConfig(settings, model)
        with fr.ThreePhaseMethod(cfg) as workflow:
            workflow.generate_matrices(view_matrices=False)
            a = workflow.calculate_sensor(
                "view_1",
                {"window_1": "blinds30"},  # blinds30 is the matrix_name
                datetime(2023, 1, 1, 12),
                800,
                100,
            )
        self.assertEqual(a.shape , (1, 1))


    def test_two_phase(self):
        time = datetime(2023, 1, 1, 12)
        dni = 800
        dhi = 100
        config = fr.WorkflowConfig.from_dict(self.cfg)
        with fr.TwoPhaseMethod(config) as workflow:
            workflow.generate_matrices()
            res = workflow.calculate_sensor("wpi", time, dni, dhi)
        self.assertEqual(res.shape , (195, 1))


    def test_three_phase(self):
        time = datetime(2023, 1, 1, 12)
        dni = 800
        dhi = 100
        config = fr.WorkflowConfig.from_dict(self.cfg)
        blind_prim = pr.Primitive(
            "void",
            "aBSDF",
            "blinds30",
            [str(self.resources_dir / "blinds30.xml"), "0", "0", "1", "."],
            [],
        )
        config.model.materials.glazing_materials = {"blinds30": blind_prim}
        with fr.ThreePhaseMethod(config) as workflow:
            workflow.generate_matrices(view_matrices=False)
            workflow.calculate_sensor(
                "wpi",
                {"upper_glass": "blinds30", "lower_glass": "blinds30"},
                time,
                dni,
                dhi,
            )
            res = workflow.calculate_edgps(
                "view1",
                {"upper_glass": "blinds30", "lower_glass": "blinds30"},
                time,
                dni,
                dhi,
            )
            res = workflow.calculate_sensor_from_wea("wpi")


    def test_eprad_threephase(self):
        """
        Simplified integration test for ThreePhaseMethod
        """
        # Use minimal parameters for faster testing
        clear_glass_path = self.resources_dir / "CLEAR_3.DAT"
        layers = [fr.window.LayerInput(clear_glass_path)]
        epmodel = fr.load_energyplus_model(ref_models["medium_office"])
        gs_simple = fr.create_glazing_system(
            name="simple_glass",
            layer_inputs=layers,
            nsamp=1,  # Minimal sampling for speed
            nproc=1,  # Single process for speed
        )
        epmodel.add_glazing_system(gs_simple)
        
        # Test basic functionality without full workflow
        self.assertEqual(gs_simple.name, "simple_glass")
        self.assertIsNotNone(epmodel.construction_complex_fenestration_state)

if __name__=="__main__":
    unittest.main()



from pathlib import Path
import unittest

from frads import geom
import numpy as np


class TestGeometry(unittest.TestCase):
    def test_merge_window(self):
        """
        0,0,7-------3,0,7   4,0,7-------8,0,7
          |           |       |           |
          |           |       |           |
          |           |       |           |
        0,0,4-------3,0,4   4,0,4-------8,0,4

        0,0,3-------3,0,3   4,0,3-------8,0,3
          |           |       |           |
          |           |       |           |
          |           |       |           |
        0,0,0-------3,0,0   4,0,0-------8,0,0
        """
        pt1 = np.array((0, 0, 0))
        pt2 = np.array((3, 0, 0))
        pt3 = np.array((4, 0, 0))
        pt4 = np.array((8, 0, 0))
        pt5 = np.array((0, 0, 3))
        pt6 = np.array((3, 0, 3))
        pt7 = np.array((4, 0, 3))
        pt8 = np.array((8, 0, 3))
        pt9 = np.array((0, 0, 4))
        pt10 = np.array((3, 0, 4))
        pt11 = np.array((4, 0, 4))
        pt12 = np.array((8, 0, 4))
        pt13 = np.array((0, 0, 7))
        pt14 = np.array((3, 0, 7))
        pt15 = np.array((4, 0, 7))
        pt16 = np.array((8, 0, 7))

        plg1 = geom.Polygon([pt1, pt5, pt6, pt2])
        plg2 = geom.Polygon([pt3, pt7, pt8, pt4])
        plg3 = geom.Polygon([pt9, pt13, pt14, pt10])
        plg4 = geom.Polygon([pt11, pt15, pt16, pt12])

        answer12 = geom.Polygon([pt1, pt5, pt8, pt4])
        answer123 = geom.Polygon([pt1, pt13, pt14, pt8, pt4])
        answer1234 = geom.Polygon([pt1, pt13, pt16, pt4])

        res12 = geom.merge_polygon([plg1, plg2])
        res123 = geom.merge_polygon([plg1, plg2, plg3])
        res1234 = geom.merge_polygon([plg1, plg2, plg3, plg4])

        self.assertEqual(res12, answer12)
        self.assertEqual(res123, answer123)
        self.assertEqual(res1234, answer1234)

    def test_trim_window(self):
        wall_width = 50
        wall_height = 3
        window_width = 50
        window_height = 1.2
        sill_height = 1.2
        south_wall = geom.Polygon([
            np.array((0, 0, 0)),
            np.array((wall_width, 0, 0)),
            np.array((wall_width, 0, wall_height)),
            np.array((0, 0, wall_height)),
        ])
        window = geom.Polygon([
            np.array((0, 0, sill_height)),
            np.array((0, 0, sill_height + window_height)),
            np.array((window_width, 0, sill_height + window_height)),
            np.array((window_width, 0, sill_height)),
        ])
        trimmed_wall = south_wall - window
        print(trimmed_wall)
        print(trimmed_wall.normal)
        self.assertEqual(trimmed_wall.normal[0], 0)
        self.assertEqual(trimmed_wall.normal[1], -1)
        self.assertEqual(trimmed_wall.normal[2], 0)

if __name__== "__main__":
    unittest.main()



from pathlib import Path
import unittest
import pywincalc as pwc
from frads.window import (
    create_glazing_system,
    Gas,
    Gap,
    LayerInput,
    OpeningDefinitions,
    Layer,
    get_layer_groups,
    load_glazing_system,
    generate_melanopic_bsdf,
)
import pyradiance as pr


class TestWindow(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.resources_dir = Path(__file__).parent / "Resources"
        glass_path = cls.resources_dir / "CLEAR_3.DAT"
        shade_path = cls.resources_dir / "2011-SA1.xml"
        blinds_path = cls.resources_dir / "igsdb_product_19732.json"
        glass_layer = LayerInput(glass_path)
        openings = OpeningDefinitions(
            left_m=0.01,
            right_m=0.005,
            top_m=0.0025,
            bottom_m=0.005,
        )
        fabric_layer = LayerInput(
            input_source=shade_path,
            openings=openings,
        )
        blinds_layer = LayerInput(
            input_source=blinds_path,
            slat_angle_deg=45,
        )
        cls.double_glaze = [glass_layer, glass_layer]
        triple_glaze = [glass_layer, glass_layer, glass_layer]
        cls.double_glaze_fabric = [glass_layer, glass_layer, fabric_layer]
        double_glaze_blinds = [glass_layer, glass_layer, blinds_layer]
        cls.single_glaze_blinds = [glass_layer, blinds_layer]
        cls.double_glaze_inner_fabric = [glass_layer, fabric_layer, glass_layer]
        cls.double_glaze_system = create_glazing_system(
            name="dgu", layer_inputs=cls.double_glaze,
            gaps=[Gap([Gas("air", 0.1), Gas("argon", 0.9)], 0.03)],
        )
        cls.triple_glaze_system = create_glazing_system(
            name="tgu", layer_inputs=triple_glaze
        )

        cls.double_glaze_fabric_system = create_glazing_system(
            name="dgu_shade", layer_inputs=cls.double_glaze_fabric
        )
        cls.double_glaze_blinds_system = create_glazing_system(
            name="dgu_blinds", layer_inputs=double_glaze_blinds, nsamp=1,
        )

    def test_save_and_load(self):
        """
        Test the save method of the GlazingSystem class.
        """
        self.double_glaze_fabric_system.save("test.json")
        self.assertTrue(Path("test.json").exists())
        gs2 = load_glazing_system("test.json")
        # os.remove("test.json")
        self.assertEqual(gs2.name, self.double_glaze_fabric_system.name)
        self.assertEqual(
            gs2.visible_back_reflectance,
            self.double_glaze_fabric_system.visible_back_reflectance,
        )

    def test_simple_glazingsystem(self):
        """
        Test the GlazingSystem class.
        Build a GlazingSystem object consisting of two layer of clear glass.

        Check the thickness of the glazing system.
        Check the order and name of the layers.
        Check the composition of the default gap.
        """

        self.assertEqual(
            self.double_glaze_system.layers[0].product_name, "Generic Clear Glass"
        )
        self.assertEqual(
            self.double_glaze_system.layers[1].product_name, "Generic Clear Glass"
        )
        self.assertEqual(self.double_glaze_system.name, "dgu")

    def test_get_layer_groups(self):
        glass_layer = Layer(
            product_name= "glass",
            thickness_m= 0.03,
            product_type= "glazing",
            conductivity= 1,
            emissivity_front= 1,
            emissivity_back= 1,
            ir_transmittance= 0,
            flipped= False,
        )
        shade_layer = Layer(
            product_name= "shade",
            thickness_m= 0.03,
            product_type= "fabric",
            conductivity= 1,
            emissivity_front= 1,
            emissivity_back= 1,
            ir_transmittance= 0,
            flipped= False,
        )
        blinds_layer = Layer(
            product_name= "blinds",
            thickness_m= 0.03,
            product_type= "blinds",
            conductivity= 1,
            emissivity_front= 1,
            emissivity_back= 1,
            ir_transmittance= 0,
            flipped= False,
        )
        group1 = get_layer_groups([glass_layer, glass_layer, shade_layer])
        group2 = get_layer_groups([shade_layer, glass_layer])
        group3 = get_layer_groups(
            [glass_layer, glass_layer, glass_layer, blinds_layer]
        )
        group4 = get_layer_groups(
            [glass_layer, shade_layer, glass_layer, blinds_layer]
        )
        group5 = get_layer_groups([glass_layer, glass_layer, glass_layer])
        self.assertEqual(group1, [("glazing", 2), ("fabric", 1)])
        self.assertEqual(group2, [("fabric", 1), ("glazing", 1)])
        self.assertEqual(group3, [("glazing", 3), ("blinds", 1)])
        self.assertEqual(
            group4, [("glazing", 1), ("fabric", 1), ("glazing", 1), ("blinds", 1)]
        )
        self.assertEqual(group5, [("glazing", 3)])

    def test_customized_gap(self):
        """
        Test the building of a customized gap.
        A 0.03 m thick gap between the two glass layers. The gap is filled with 90% argon and 10% air.

        Check the thickness of the glazing system.
        Check the order and composition of the gap.
        """
        self.assertEqual(self.double_glaze_system.gaps[0].gas[0].gas, "air")
        self.assertEqual(self.double_glaze_system.gaps[0].gas[0].ratio, 0.1)
        self.assertEqual(self.double_glaze_system.gaps[0].gas[1].gas, "argon")
        self.assertEqual(self.double_glaze_system.gaps[0].gas[1].ratio, 0.9)
        self.assertEqual(self.double_glaze_system.gaps[0].thickness_m, 0.03)

    def test_venetian_blinds(self):
        """
        Test GlazingSystem object with multiple layers of glazing and shading and more than one customized gap.

        Check the thickness of the glazing system.
        Check the order of the layers.
        Check the order and composition of the gaps.
        """

        self.assertEqual(self.double_glaze_blinds_system.layers[0].product_name, "Generic Clear Glass")
        self.assertEqual(self.double_glaze_blinds_system.layers[0].thickness_m, 0.003048)
        self.assertEqual(self.double_glaze_blinds_system.layers[2].product_name, "ODL Espresso Blind 14.8mm Slat")
        self.assertEqual(self.double_glaze_blinds_system.layers[2].thickness_m, 0.010465180361560904)

        self.assertEqual(self.double_glaze_blinds_system.gaps[0].gas[0].gas, "air")
        self.assertEqual(self.double_glaze_blinds_system.gaps[0].gas[0].ratio, 1)
        self.assertEqual(self.double_glaze_blinds_system.gaps[0].thickness_m, 0.0127)

        self.assertEqual(self.double_glaze_blinds_system.thickness, 0.04196118036156091)

    def test_inner_shade_gap(self):
        """
        Test left, right, top and bottom opening multipliers for inner shade with customized gap.

        Check gaps thickness
        Check the order of the gaps
        Check the thickness of the glazing system.
        Check the opening multipliers of the shading layer
        """

        gs = create_glazing_system(
            name="dgu_inner_shade",
            layer_inputs=self.double_glaze_inner_fabric,
            gaps=[Gap([Gas("air", 1)], 0.01), Gap([Gas("air", 1)], 0.005)],
        )

        self.assertEqual(gs.gaps[0].gas[0].gas, "air")
        self.assertEqual(gs.gaps[0].gas[0].ratio, 1)
        self.assertEqual(gs.gaps[0].thickness_m, 0.01)

        self.assertEqual(gs.gaps[1].gas[0].gas, "air")
        self.assertEqual(gs.gaps[1].gas[0].ratio, 1)
        self.assertEqual(gs.gaps[1].thickness_m, 0.005)

        self.assertEqual(gs.thickness, 0.022096)

        self.assertEqual(gs.layers[1].opening_multipliers.left, 1)
        self.assertEqual(gs.layers[1].opening_multipliers.right, 1)
        self.assertEqual(gs.layers[1].opening_multipliers.top, 0.5)
        self.assertEqual(gs.layers[1].opening_multipliers.bottom, 1)
        self.assertEqual(gs.layers[1].opening_multipliers.front, 0.05)

class TestBSDFGeneration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.resources_dir = Path(__file__).parent / "Resources"

    def test_melanopic_bsdf(self):
        product_data = pwc.parse_json_file(str(self.resources_dir / "igsdb_product_7406.json"))
        NM_PER_MM = 1e3
        layer = Layer(
            product_name= "glass",
            thickness_m= 0.03,
            product_type= "glazing",
            conductivity= 1,
            emissivity_front= 1,
            emissivity_back= 1,
            ir_transmittance= 0,
            flipped= False,
            spectral_data= {
                int(round(d.wavelength * NM_PER_MM)): (
                    d.direct_component.transmittance_front,
                    d.direct_component.reflectance_front,
                    d.direct_component.reflectance_back,
                )
                for d in product_data.measurements
            }
        )
        layer2 = Layer(
            product_name= "blinds",
            thickness_m= 0.03,
            product_type= "blinds",
            conductivity= 1,
            emissivity_front= 1,
            emissivity_back= 1,
            ir_transmittance= 0,
            flipped= False,
            shading_material=pr.ShadingMaterial(0.5, 0, 0),
            slat_width_m = 0.0160,
            slat_spacing_m = 0.0120,
            slat_thickness_m = 0.0006,
            slat_curve_m = 0.0,
            slat_angle_deg = 45.0,
            slat_conductivity = 160.00,
            nslats = 10,
        )
        result = generate_melanopic_bsdf(layers=[layer, layer2], gaps=[Gap([Gas("air", 0.1), Gas("argon", 0.9)], 0.03)],nsamp=1)

if __name__ == "__main__":
    unittest.main()



"""
"""
from pathlib import Path
from frads import epmodel_to_radmodel
from frads.eplus import load_energyplus_model
import unittest

class TestEp2Rad(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Setup the resources directory - replace with your actual resources path
        cls.resources_dir = Path(__file__).parent / "Resources"

    def test_cfscase(self):
        epjson1 = self.resources_dir / "CmplxGlz_SingleZone_DoubleClearAir.epJSON"
        epmodel = load_energyplus_model(epjson1)
        rad_model = epmodel_to_radmodel(epmodel)
        assert (
            rad_model["Room 102"]["model"]["sensors"]["Room102 Floor"] is not None
        )

    def test_refcase(self):
        epjson2 = self.resources_dir / "RefBldgMediumOfficeNew2004_Chicago_epJSON.epJSON"
        epmodel = load_energyplus_model(epjson2)
        rad_model = epmodel_to_radmodel(epmodel)
        assert (
            rad_model["Perimeter_bot_ZN_1"]["model"]["sensors"][
                "Perimeter_bot_ZN_1_Floor"
            ]
            is not None
        )

if __name__ == "__main__":
    unittest.main()



"""Matrix generation and manipulation for Radiance simulations.

This module provides classes and functions for creating sender and receiver objects,
generating view factor and daylight matrices using Radiance tools (rfluxmtx, rcontrib),
and performing matrix operations for lighting simulations.

Key Components:
    Sender Classes:
        - SensorSender: Point sensors for illuminance calculations
        - ViewSender: Camera views for luminance images
        - SurfaceSender: Surface-based senders with sampling basis

    Receiver Classes:
        - SurfaceReceiver: Surface-based receivers with sampling basis
        - SkyReceiver: Sky dome receiver for daylight calculations
        - SunReceiver: Solar disc receiver for direct sun calculations

    Matrix Classes:
        - Matrix: General matrix for view factor and daylight calculations
        - SunMatrix: Specialized matrix for sun-only calculations

The module supports various sampling bases (Klems, Reinhart-Tregenza, Shirley-Chiu)
and can generate both dense and sparse matrices for memory efficiency.
"""

import gc
import logging
import os
from pathlib import Path
import re
from tempfile import TemporaryDirectory

from frads.geom import parse_polygon, polygon_primitive, Polygon
from frads.utils import random_string
import numpy as np
import pyradiance as pr
from scipy.sparse import csr_matrix, lil_matrix


logger: logging.Logger = logging.getLogger("frads.matrix")


BASIS_DIMENSION = {
    "kf": 145,
    "kq": 41,
    "kh": 86,
    "r1": 145,
    "r2": 577,
    "r4": 2305,
    "r6": 5185,
    "sc2": 16,
    "sc4": 256,
    "sc5": 1024,
    "sc6": 4096,
    "sc7": 16384,
    "u": 1,
}


class SensorSender:
    """Sender object representing a collection of point sensors.

    This class creates a sender object from a list of point sensors, where each
    sensor is defined by position and direction vectors. Used for illuminance
    calculations at specific points in space.

    Attributes:
        sensors: List of sensors, each containing 6 float values [x, y, z, dx, dy, dz]
            representing position (x,y,z) and direction (dx,dy,dz) vectors.
        content: Sensor data encoded as bytes for Radiance input.
        yres: Number of unique sensor locations (before ray multiplication).
    """

    def __init__(self, sensors: list[list[float]], ray_count: int = 1):
        """Initialize a sensor sender object.

        Args:
            sensors: List of sensors, each containing 6 float values [x, y, z, dx, dy, dz]
                representing position and direction vectors.
            ray_count: Number of rays to generate per sensor for Monte Carlo sampling.
                Higher values improve accuracy but increase computation time.
        """
        self.sensors = [i for i in sensors for _ in range(ray_count)]
        self.content = (
            os.linesep.join([" ".join(map(str, li)) for li in self.sensors])
            + os.linesep
        ).encode()
        self.yres = len(sensors)

    def __eq__(self, other):
        return self.content == other.content


class ViewSender:
    """Sender object representing a camera view for luminance calculations.

    This class creates a sender object from a camera view definition, generating
    rays for each pixel in the view. Used for creating luminance images and
    view-based daylight analysis.

    Attributes:
        view: Pyradiance View object defining camera parameters (position, direction, etc.).
        content: View rays encoded as bytes for Radiance input.
        xres: Horizontal resolution (number of pixels in x-direction).
        yres: Vertical resolution (number of pixels in y-direction).
    """

    def __init__(
        self, view: pr.View, ray_count: int = 1, xres: int = 800, yres: int = 800
    ):
        """Initialize a view sender object.

        Args:
            view: Pyradiance View object defining camera parameters.
            ray_count: Number of rays per pixel for anti-aliasing and Monte Carlo sampling.
            xres: Horizontal resolution in pixels.
            yres: Vertical resolution in pixels.
        """
        self.view = view
        self.xres = xres
        self.yres = yres
        view_args = pr.get_view_args(view)
        res_eval = pr.vwrays(
            view=view_args, xres=xres, yres=yres, dimensions=True
        ).split()
        new_xres, new_yres = int(res_eval[1]), int(res_eval[3])
        vwrays_proc = pr.vwrays(
            view=view_args,
            outform="f",
            xres=new_xres,
            yres=new_yres,
            ray_count=ray_count,
            pixel_jitter=0.7,
        )
        if view.type == "a":
            ray_flush_exp = (
                f"DIM:{xres};CNT:{ray_count};"
                "pn=(recno-1)/CNT+.5;"
                "frac(x):x-floor(x);"
                "xpos=frac(pn/DIM);ypos=pn/(DIM*DIM);"
                "incir=if(.25-(xpos-.5)*(xpos-.5)-(ypos-.5)*(ypos-.5),1,0);"
                "$1=$1;$2=$2;$3=$3;$4=$4*incir;$5=$5*incir;$6=$6*incir;"
            )
            flushed_rays = pr.rcalc(
                vwrays_proc, inform="f", incount=6, outform="f", expr=ray_flush_exp
            )
            vrays = flushed_rays
        else:
            vrays = vwrays_proc
        self.content = vrays

    def __eq__(self, other):
        return (
            self.view == other.view
            and self.xres == other.xres
            and self.yres == other.yres
        )


class SurfaceSender:
    """Sender object representing surfaces with directional sampling.

    This class creates a sender object from surface primitives using a specified
    sampling basis. Used for calculating inter-surface view factors and daylight
    transmission through surfaces like windows.

    Attributes:
        surfaces: List of surface primitives (polygons or rings) to use as senders.
        basis: Sampling basis string (e.g., 'kf', 'r4', 'sc6') defining directional
            discretization for the surface.
        content: Surface data encoded as string for Radiance rfluxmtx input.
    """

    def __init__(
        self,
        surfaces: list[pr.Primitive],
        basis: str,
        left_hand: bool = False,
        offset: float | None = None,
    ):
        """Initialize a surface sender object.

        Args:
            surfaces: List of surface primitives (polygons or rings) to use as senders.
            basis: Sampling basis string defining directional discretization
                (e.g., 'kf' for Klems full, 'r4' for Reinhart 4-division).
            left_hand: Whether to use left-hand coordinate system for basis orientation.
            offset: Distance to offset the sender surface along its normal vector.
                Useful for avoiding self-intersection in calculations.
        """
        self.surfaces = [s for s in surfaces if s.ptype in ("polygon", "ring")]
        self.basis = basis
        self.content = rfluxmtx_markup(
            self.surfaces, basis, left_hand=left_hand, offset=offset
        )

    def __eq__(self, other):
        return self.content == other.content


class Receiver:
    """Base class for all receiver objects in matrix calculations.

    This abstract base class defines the common interface for receiver objects
    that collect light from senders. All receiver types inherit from this class
    and implement specific encoding for different receiver geometries.

    Attributes:
        basis: Sampling basis string defining directional discretization for the receiver.
        content: Receiver data encoded as string for Radiance input.
    """

    def __init__(self, basis: str):
        """Initialize a receiver object.

        Args:
            basis: Sampling basis string (e.g., 'kf', 'r4', 'sc6', 'u') defining
                the directional discretization scheme for the receiver.

        Raises:
            ValueError: If the basis string format is invalid.
        """
        check_index = 0
        if basis[0] == "-":
            check_index = 1
        if basis[check_index] not in ("u", "k", "r", "s"):
            raise ValueError("Invalid basis")
        if basis[check_index] == "k" and basis[check_index + 1] not in ("f", "q", "h"):
            raise ValueError("Invalid Klems basis", basis)
        if basis[check_index] == "r" and not basis[check_index + 1].isdigit():
            raise ValueError("Invalid Reinhart/Treganza basis", basis)
        if basis[check_index] == "s" and not basis[check_index + 2].isdigit():
            raise ValueError("Invalid Shirley-Chiu basis", basis)
        self.basis = basis
        self.content = ""


class SurfaceReceiver(Receiver):
    """Receiver object representing surfaces that collect light.

    This class creates a receiver object from surface primitives using a specified
    sampling basis. Used for calculating how much light surfaces receive from
    various sources in daylight and artificial lighting simulations.

    Attributes:
        surfaces: List of surface primitives (polygons or rings) acting as receivers.
        basis: Sampling basis string defining directional discretization.
        content: Surface data encoded as string for Radiance rfluxmtx input.
    """

    def __init__(
        self,
        surfaces: list[pr.Primitive],
        basis: str,
        left_hand: bool = False,
        offset: float | None = None,
        source: str = "glow",
        out: str | None = None,
    ):
        """Initialize a surface receiver object.

        Args:
            surfaces: List of surface primitives (polygons or rings) to use as receivers.
            basis: Sampling basis string defining directional discretization.
            left_hand: Whether to use left-hand coordinate system for basis orientation.
            offset: Distance to offset the receiver surface along its normal vector.
            source: Light source type for the receiver ('glow' or 'light').
            out: Optional output file path for matrix results.

        Raises:
            ValueError: If surfaces are not primitives or contain invalid types.
        """
        super().__init__(basis)
        if not isinstance(surfaces[0], pr.Primitive):
            raise ValueError("Surface must be a primitive", surfaces)
        self.surfaces = [s for s in surfaces if s.ptype in ("polygon", "ring")]
        self.content = rfluxmtx_markup(
            self.surfaces,
            basis=basis,
            left_hand=left_hand,
            offset=offset,
            source=source,
            out=out,
        )


class SkyReceiver(Receiver):
    """Receiver object representing the sky dome for daylight calculations.

    This class creates a receiver object that represents the entire sky dome,
    including both sky and ground hemispheres. Used in daylight matrix calculations
    to capture contributions from all sky directions.

    Attributes:
        basis: Sampling basis string defining sky patch discretization.
        content: Sky dome definition encoded as string for Radiance input.
    """

    def __init__(self, basis: str, out: str | None = None):
        """Initialize a sky receiver object.

        Args:
            basis: Sampling basis string defining sky patch discretization
                (e.g., 'kf' for Klems full, 'r4' for Reinhart 4-division).
            out: Optional output file path for matrix results.
        """
        super().__init__(basis)
        self.content = ""
        if out is not None:
            self.content += f"#@rfluxmtx o={out}\n"
        self.content += (
            f"#@rfluxmtx h=u\n"
            "void glow groundglow 0 0 4 1 1 1 0\n"
            "groundglow source groundplane 0 0 4 0 0 -1 180\n"
            f"#@rfluxmtx h={basis} u=+Y\n"
            "void glow skyglow 0 0 4 1 1 1 0\n"
            "skyglow source skydome 0 0 4 0 0 1 180\n"
        )


class SunReceiver(Receiver):
    """Receiver object representing discrete solar positions for direct sun calculations.

    This class creates a receiver object that represents the sun at discrete positions
    based on a Reinhart-Tregenza sky division. The number of active sun positions
    can be reduced based on annual solar data or window visibility constraints.

    Attributes:
        basis: Sampling basis string (must be Reinhart-Tregenza format like 'r4').
        content: Sun positions encoded as Radiance light sources.
        modifiers: List of sun modifier names for active solar positions.
    """

    def __init__(
        self,
        basis: str,
        sun_matrix: None | np.ndarray = None,
        window_normals: None | list[np.ndarray] = None,
        full_mod: bool = False,
    ):
        """Initialize a sun receiver object.

        Args:
            basis: Sampling basis string (must be Reinhart-Tregenza format like 'r4').
            sun_matrix: Optional annual sun matrix to filter out zero-contribution
                solar positions. Shape should be (time_steps, sun_positions, components).
            window_normals: Optional list of window normal vectors to filter out
                sun positions not visible through windows.
            full_mod: If True, include all sun modifiers regardless of filtering.
                If False, only include modifiers for active sun positions.

        Raises:
            ValueError: If basis is not a valid Reinhart-Tregenza format.
        """
        super().__init__(basis)
        if not basis.startswith("r") and not basis[-1].isdigit():
            raise ValueError("Invalid Reinhart/Treganza basis", basis)
        mf = int(basis[-1])
        nbins = 144 * mf**2 + 1
        res = pr.rcalc(
            inp=pr.cnt(nbins),
            outform="f",
            expr=f"MF:{mf};Rbin=recno;$1=Dx;$2=Dy;$3=Dz",
            source="reinsrc.cal",
        )
        sundirs = np.frombuffer(res, dtype=np.single).reshape(nbins, 3)
        sunvals = np.ones(BASIS_DIMENSION[basis])
        if sun_matrix is not None:
            sunvals[np.sum(sun_matrix[1:, :, 0], axis=1) == 0] = 0
        if window_normals is not None:
            sunvals[np.dot(sundirs, np.array(window_normals).T).flatten() >= 0] = 0
        self.content = "\n".join(
            f"void light sol{i} 0 0 3 {d} {d} {d} sol{i} source sun 0 0 4 {sundirs[i][0]:.6g} {sundirs[i][1]:.6g} {sundirs[i][2]:.6g} 0.533"
            for i, d in enumerate(sunvals)
        )
        if full_mod:
            self.modifiers = [f"sol{i}" for i in range(nbins)]
        else:
            self.modifiers = [f"sol{i}" for i in np.where(sunvals > 0)[0]]


class Matrix:
    """General matrix object for daylight and view factor calculations.

    This class represents a transfer matrix between senders and receivers, encoding
    how light travels from source points/surfaces to destination points/surfaces.
    Used for daylight simulations, view factor calculations, and lighting analysis.

    Attributes:
        sender: Sender object (sensors, view, or surface) that emits light.
        receivers: List of receiver objects that collect light.
        array: Numpy array or list of arrays storing the matrix data with shape
            (nrows, ncols, ncomp) where ncomp is typically 3 for RGB.
        ncols: Number of columns (receiver basis size) or list of sizes for multiple receivers.
        nrows: Number of rows (sender basis size or sensor count).
        dtype: Matrix data type ('d' for double, 'f' for float).
        ncomp: Number of color components (typically 3 for RGB).
        octree: Path to octree file used for ray tracing acceleration.
        surfaces: List of environment surface primitives for the scene.
    """

    def __init__(
        self,
        sender: SensorSender | ViewSender | SurfaceSender,
        receivers: list[SkyReceiver] | list[SurfaceReceiver] | list[SunReceiver],
        octree: None | str = None,
        surfaces: None | list[pr.Primitive] = None,
    ):
        """Initialize a matrix object.

        Args:
            sender: Sender object (SensorSender, ViewSender, or SurfaceSender).
            receivers: List of receiver objects (SurfaceReceiver, SkyReceiver, or SunReceiver).
                Multiple receivers are only supported for SurfaceReceiver types.
            octree: Optional path to octree file for ray tracing.
            surfaces: Optional list of environment surface primitives for the scene.

        Raises:
            ValueError: If sender/receiver types are invalid or incompatible.
        """
        if not isinstance(sender, (SensorSender, ViewSender, SurfaceSender)):
            raise ValueError(
                "Sender must be a SensorSender, ViewSender, or SurfaceSender"
            )
        if not isinstance(receivers[0], (SurfaceReceiver, SkyReceiver, SunReceiver)):
            raise ValueError(
                "Receiver must be a SurfaceReceiver, SkyReceiver, or SunReceiver"
            )
        if len(receivers) > 1:
            if not all(isinstance(r, SurfaceReceiver) for r in receivers):
                raise ValueError("All receivers must be SurfaceReceivers")
        self.sender = sender
        self.receivers = receivers
        self.array = None
        self.ncols = None
        self.nrows: int
        self.dtype = "d"
        self.ncomp = 3
        if isinstance(sender, SensorSender):
            self.nrows = sender.yres
        elif isinstance(sender, SurfaceSender):
            self.nrows = BASIS_DIMENSION[sender.basis]
        elif isinstance(sender, ViewSender):
            self.nrows = sender.yres * sender.xres
        if isinstance(receivers[0], SurfaceReceiver):
            self.ncols = [BASIS_DIMENSION[r.basis] for r in receivers]
        elif isinstance(receivers[0], SkyReceiver):
            self.ncols = BASIS_DIMENSION[receivers[0].basis] + 1
        self.octree = octree
        self.surfaces = surfaces

    def generate(
        self,
        params: list[str],
        nproc: int = 1,
        sparse: bool = False,
        to_file: bool = False,
        memmap: bool = False,
    ) -> None:
        """Generate the matrix using Radiance rfluxmtx tool.

        Args:
            params: List of rfluxmtx command-line parameters (e.g., ['-ab', '3', '-ad', '1000']).
            nproc: Number of parallel processes to use for calculation.
            sparse: If True, convert result to sparse matrix format for memory efficiency.
            to_file: If True, write matrix directly to file without storing in memory.
                Useful for very large matrices that exceed available RAM.
            memmap: If True, use memory mapping to store matrix data on disk.
                Allows handling of matrices larger than available RAM.
        """
        surface_file = None
        rays = None
        params.append("-n")
        params.append(f"{nproc}")
        logger.info("Generating matrix...")
        if logger.getEffectiveLevel() > 20:
            params.append("-w")
        if not isinstance(self.sender, SurfaceSender):
            rays = self.sender.content
            params.append("-y")
            params.append(str(self.sender.yres))
            if isinstance(self.sender, SensorSender):
                params.append("-I")
                params.append("-fad")
            elif isinstance(self.sender, ViewSender):
                params.append("-x")
                params.append(str(self.sender.xres))
                params.append("-ffd")
        with TemporaryDirectory() as tmpdir:
            receiver_file = os.path.join(tmpdir, "receiver")
            with open(receiver_file, "w") as f:
                [f.write(r.content) for r in self.receivers]
            env_file = None
            if self.surfaces is not None:
                env_file = os.path.join(tmpdir, "scene")
                with open(env_file, "wb") as f:
                    f.write(b" ".join(s.bytes for s in self.surfaces))
                env_file = [env_file]
            if isinstance(self.sender, SurfaceSender):
                params.append("-ffd")
                surface_file = os.path.join(tmpdir, "surface")
                with open(surface_file, "w") as f:
                    f.write(self.sender.content)
            matrix = pr.rfluxmtx(
                receiver_file,
                surface=surface_file,
                rays=rays,
                params=params,
                octree=self.octree,
                scene=env_file,
            )
        if not to_file:
            _ncols = sum(self.ncols) if isinstance(self.ncols, list) else self.ncols
            _array = load_binary_matrix(
                matrix,
                self.nrows,
                _ncols,
                self.ncomp,
                self.dtype,
                header=True,
            )
            del matrix
            gc.collect()
            if memmap:
                _dtype = np.float64 if self.dtype.startswith("d") else np.float32
                self.array = np.memmap(
                    f"{random_string(5)}.dat",
                    dtype=_dtype,
                    shape=(self.nrows, _ncols, self.ncomp),
                    order="F",
                    mode="w+",
                )
                self.array[:] = _array
                self.array.flush()
            else:
                self.array = _array
            del _array
            gc.collect()
            # If multiple receivers, split the array horizontally
            if isinstance(self.ncols, list):
                self.array = np.hsplit(self.array, np.cumsum(self.ncols)[:-1])
                if sparse:
                    self.array = np.array(
                        [
                            np.array(
                                (
                                    csr_matrix(a[:, :, 0]),
                                    csr_matrix(a[:, :, 1]),
                                    csr_matrix(a[:, :, 2]),
                                )
                            )
                            for a in self.array
                        ]
                    )
            elif sparse:
                self.array = np.array(
                    (
                        csr_matrix(self.array[:, :, 0]),
                        csr_matrix(self.array[:, :, 1]),
                        csr_matrix(self.array[:, :, 2]),
                    )
                )


class SunMatrix(Matrix):
    """Specialized matrix object for direct solar radiation calculations.

    This class extends the base Matrix class to handle sun-only calculations using
    the Radiance rcontrib tool. It's optimized for sparse matrices since most
    solar positions contribute zero radiation at any given time.

    Attributes:
        sender: Sender object (SensorSender or ViewSender only).
        receiver: SunReceiver object representing discrete solar positions.
        octree: Path to octree file for ray tracing acceleration.
        surfaces: List of environment surface file paths.
        array: Sparse matrix array storing sun contribution data.
        nrows: Number of rows (sender basis size or sensor count).
        ncols: Number of columns (solar position count).
        dtype: Matrix data type ('d' for double, 'f' for float).
        ncomp: Number of color components (typically 3 for RGB).
    """

    def __init__(
        self,
        sender: SensorSender | ViewSender,
        receiver: SunReceiver,
        octree: str | None,
        surfaces: list[str] | None = None,
    ):
        """Initialize a sun matrix object.

        Args:
            sender: Sender object (SensorSender or ViewSender only).
            receiver: SunReceiver object representing discrete solar positions.
            octree: Optional path to octree file for ray tracing acceleration.
            surfaces: Optional list of environment surface file paths.

        Raises:
            TypeError: If sender is a SurfaceSender (not supported for sun matrices).
        """
        if isinstance(sender, SurfaceSender):
            raise TypeError("SurfaceSender cannot be used with SunMatrix")
        super().__init__(sender, [receiver], octree, surfaces=surfaces)
        self.surfaces = [] if surfaces is None else surfaces
        self.receiver = receiver
        self.ncols = BASIS_DIMENSION[receiver.basis] + 1

    def generate(
        self,
        parameters: list[str],
        nproc: int = 1,
        radmtx: bool = False,
        sparse: bool = False,
    ) -> None | bytes:
        """Generate the sun matrix using Radiance rcontrib tool.

        Args:
            parameters: List of rcontrib command-line parameters (e.g., ['-ab', '1', '-ad', '512']).
            nproc: Number of parallel processes to use for calculation.
            radmtx: If True, return raw matrix bytes instead of processing into numpy array.
            sparse: If True, store result as sparse matrix format (recommended for sun matrices
                due to their inherently sparse nature).

        Returns:
            Raw matrix bytes if radmtx=True, otherwise None (matrix stored in self.array).

        Raises:
            TypeError: If receiver is not a SunReceiver object.
        """
        if not isinstance(self.receiver, SunReceiver):
            raise TypeError("SunMatrix must have a SunReceiver")
        xres, yres = None, None
        inform = "a"
        parameters.append("-n")
        parameters.append(f"{nproc}")
        parameters.append("-h")
        if logger.getEffectiveLevel() > 20:
            parameters.append("-w")
        logger.info("Generating matrix...")
        with TemporaryDirectory() as tmpdir:
            octree_file = os.path.join(tmpdir, "octree")
            receiver_file = os.path.join(tmpdir, "receiver")
            modifier_file = os.path.join(tmpdir, "modifier")
            with open(modifier_file, "w") as f:
                f.write("\n".join(self.receiver.modifiers))
            with open(receiver_file, "w") as f:
                f.write(self.receiver.content)
            with open(octree_file, "wb") as f:
                f.write(pr.oconv(receiver_file, *self.surfaces, octree=self.octree))
            if isinstance(self.sender, SensorSender):
                parameters.append("-I+")
                yres = self.sender.yres
            elif isinstance(self.sender, ViewSender):
                inform = "f"
                xres = self.sender.xres
                yres = self.sender.yres
            modifier = pr.RcModifier()
            modifier.modifier_path = modifier_file
            modifier.xres = xres
            modifier.yres = yres
            matrix = pr.rcontrib(
                self.sender.content,
                octree_file,
                [modifier],
                inform=inform,
                outform="d",
                params=parameters,
            )
        if radmtx:
            return matrix
        _array = load_binary_matrix(
            matrix,
            self.nrows,
            min(self.ncols, len(self.receiver.modifiers)),
            self.ncomp,
            self.dtype,
            header=False,
        )
        # Convert to sparse matrix
        sparse_r = lil_matrix(_array[:, :, 0])
        sparse_g = lil_matrix(_array[:, :, 1])
        sparse_b = lil_matrix(_array[:, :, 2])
        del matrix, _array
        gc.collect()
        # Fill the matrix back up to full basis size if culled
        if len(self.receiver.modifiers) < self.ncols:
            indices = [int(m.lstrip("sol")) for m in self.receiver.modifiers]
            padded_sparse_r = lil_matrix((self.nrows, self.ncols))
            padded_sparse_g = lil_matrix((self.nrows, self.ncols))
            padded_sparse_b = lil_matrix((self.nrows, self.ncols))
            padded_sparse_r[:, indices] = sparse_r
            padded_sparse_g[:, indices] = sparse_g
            padded_sparse_b[:, indices] = sparse_b
            self.array = np.array(
                (
                    padded_sparse_r.tocsr(),
                    padded_sparse_g.tocsr(),
                    padded_sparse_b.tocsr(),
                )
            )
        else:
            self.array = np.array(
                (sparse_r.tocsr(), sparse_g.tocsr(), sparse_b.tocsr())
            )


def parse_rad_header(header_str: str) -> tuple[int, int, int, str]:
    """Parse a Radiance matrix file header to extract matrix dimensions and format.

    Args:
        header_str: Header string from a Radiance matrix file containing metadata.

    Returns:
        Tuple containing (nrows, ncols, ncomp, dtype) where:
            - nrows: Number of matrix rows
            - ncols: Number of matrix columns
            - ncomp: Number of color components (typically 3 for RGB)
            - dtype: Data type string (e.g., 'double', 'float')

    Raises:
        ValueError: If any required header entries (NROWS, NCOLS, NCOMP, FORMAT) are missing.
    """
    compiled = re.compile(
        r" NROWS=(.*) | NCOLS=(.*) | NCOMP=(.*) | FORMAT=(.*) ", flags=re.X
    )
    matches = compiled.findall(header_str)
    if len(matches) != 4:
        raise ValueError("Can't find one of the header entries.")
    nrow = int([mat[0] for mat in matches if mat[0] != ""][0])
    ncol = int([mat[1] for mat in matches if mat[1] != ""][0])
    ncomp = int([mat[2] for mat in matches if mat[2] != ""][0])
    dtype = [mat[3] for mat in matches if mat[3] != ""][0].strip()
    return nrow, ncol, ncomp, dtype


def load_matrix(file: bytes | str | Path, dtype: str = "float") -> np.ndarray:
    """Load a Radiance matrix file into a numpy array.

    Args:
        file: Path to Radiance matrix file or raw matrix bytes.
        dtype: Data type for the output array ('float' or 'double').

    Returns:
        Numpy array with shape (nrows, ncols, ncomp) containing the matrix data.
    """
    npdtype = np.double if dtype.startswith("d") else np.single
    mtx = pr.rmtxop(file, outform=dtype[0].lower())
    nrows, ncols, ncomp, _ = parse_rad_header(pr.getinfo(mtx).decode())
    return np.frombuffer(pr.getinfo(mtx, strip_header=True), dtype=npdtype).reshape(
        nrows, ncols, ncomp
    )


def load_binary_matrix(
    buffer: bytes, nrows: int, ncols: int, ncomp: int, dtype: str, header: bool = False
) -> np.ndarray:
    """Load a binary matrix buffer into a numpy array.

    Args:
        buffer: Raw binary data containing the matrix.
        nrows: Number of matrix rows.
        ncols: Number of matrix columns.
        ncomp: Number of color components (typically 3 for RGB).
        dtype: Data type string ('d' for double, 'f' for float).
        header: If True, strip Radiance header from buffer before parsing.

    Returns:
        Numpy array with shape (nrows, ncols, ncomp) containing the matrix data.
    """
    npdtype = np.double if dtype.startswith("d") else np.single
    if header:
        buffer = pr.getinfo(buffer, strip_header=True)
    return np.frombuffer(buffer, dtype=npdtype).reshape(nrows, ncols, ncomp)


def matrix_multiply_rgb(
    *mtx: np.ndarray, weights: list[float] | None = None
) -> np.ndarray:
    """Multiply RGB matrices using optimized matrix multiplication.

    Performs matrix multiplication on RGB matrices by multiplying each color
    channel separately. Uses numpy's multi_dot for optimal multiplication order.

    Args:
        *mtx: Variable number of matrices to multiply, each with shape (..., ..., 3).
        weights: Optional RGB weights [r_weight, g_weight, b_weight] to combine
            color channels into a single weighted result.

    Returns:
        Result matrix with same shape as input matrices. If weights provided,
        returns single-channel weighted combination, otherwise returns RGB matrix.

    Raises:
        ValueError: If weights list doesn't contain exactly 3 values.
    """
    resr = np.linalg.multi_dot([m[:, :, 0] for m in mtx])
    resg = np.linalg.multi_dot([m[:, :, 1] for m in mtx])
    resb = np.linalg.multi_dot([m[:, :, 2] for m in mtx])
    if weights:
        if len(weights) != 3:
            raise ValueError("Weights should have 3 values")
        return resr * weights[0] + resg * weights[1] + resb * weights[2]
    return np.dstack((resr, resg, resb))


def sparse_matrix_multiply_rgb_vtds(
    vmx: np.ndarray,
    tmx: np.ndarray,
    dmx: np.ndarray,
    smx: np.ndarray,
    weights: list[float] | None = None,
) -> np.ndarray:
    """Multiply view, transmission, daylight, and sky matrices for three-phase method.

    Performs the matrix multiplication sequence V × T × D × S for three-phase daylight
    calculations, handling sparse matrices efficiently for memory optimization.

    Args:
        vmx: View matrix (sparse) with shape (3, view_points, surface_patches).
        tmx: Transmission matrix (dense) with shape (surface_patches, sky_patches, 3).
        dmx: Daylight matrix (sparse) with shape (3, sky_patches, sky_directions).
        smx: Sky matrix (sparse) with shape (3, sky_directions, time_steps).
        weights: Optional RGB weights to combine color channels into weighted result.

    Returns:
        Dense numpy array with final illuminance/luminance values. Shape depends on
        whether weights are provided (single channel vs RGB).

    Raises:
        ValueError: If weights length doesn't match number of matrix channels.
    """
    if weights is not None:
        if len(weights) != vmx.shape[0]:
            raise ValueError("Mismatch between weights and matrix channels")
    # Looping through each of the RGB channels
    _res = []
    for c in range(vmx.shape[0]):
        td_mtx = np.dot(csr_matrix(tmx[:, :, c]), dmx[c])
        tds_mtx = np.dot(td_mtx, smx[c])
        _res.append(np.dot(vmx[c], tds_mtx))
    if weights is not None:
        res = np.zeros((vmx[0].shape[0], smx[0].shape[1]))
        for i, w in enumerate(weights):
            res += _res[i] * w
    else:
        res = np.dstack(_res)
    return res


def to_sparse_matrix3(mtx: np.ndarray, mtype: str = "csr") -> np.ndarray:
    """Convert a three-channel RGB matrix to sparse matrix format.

    Args:
        mtx: Dense matrix with shape (..., ..., 3) representing RGB data.
        mtype: Sparse matrix type ('csr' for compressed sparse row,
            'lil' for list of lists format).

    Returns:
        Numpy array containing 3 sparse matrices, one for each RGB channel.

    Raises:
        ValueError: If matrix doesn't have exactly 3 channels.
    """
    sparser = {
        "csr": csr_matrix,
        "lil": lil_matrix,
    }
    if mtx.shape[2] != 3:
        raise ValueError("Matrix must have 3 channels")
    return np.array(
        (
            sparser[mtype](mtx[:, :, 0]),
            sparser[mtype](mtx[:, :, 1]),
            sparser[mtype](mtx[:, :, 2]),
        )
    )


def rfluxmtx_markup(
    surfaces: list[pr.Primitive],
    basis: str,
    left_hand: bool = False,
    offset: float | None = None,
    source: str = "glow",
    out: str | None = None,
) -> str:
    """Generate rfluxmtx markup for surface primitives.

    Creates properly formatted Radiance scene description with rfluxmtx directives
    for matrix generation. Handles basis orientation, surface offsetting, and
    source material assignment.

    Args:
        surfaces: List of surface primitives (polygons or rings) to mark up.
        basis: Sampling basis string (e.g., 'kf', 'r4', 'sc6') for directional discretization.
        left_hand: If True, use left-hand coordinate system for basis orientation.
        offset: Distance to offset surfaces along their normal vectors.
        source: Light source material type ('glow' or 'light').
        out: Optional output file path for matrix results.

    Returns:
        String containing formatted Radiance scene with rfluxmtx directives.

    Raises:
        ValueError: If multiple surface modifiers found or invalid source type.
    """
    modifier_set = {p.modifier for p in surfaces}
    source_modifier = f"rflx{surfaces[0].modifier}{random_string(10)}"
    if left_hand:
        basis = "-" + basis
    if source not in ("glow", "light"):
        raise ValueError("Invalid source")
    primitives = [p for p in surfaces if p.ptype in ("polygon", "ring")]
    surface_normal = np.zeros(3)
    for primitive in primitives:
        polygon = parse_polygon(primitive)
        surface_normal += polygon.area
    sampling_direction = surface_normal * (1 / len(primitives))
    sampling_direction = sampling_direction / np.linalg.norm(sampling_direction)
    if np.array_equal(sampling_direction, np.array([0, 0, 1])):
        up_vector = np.array([0, 1, 0])
    elif np.array_equal(sampling_direction, np.array([0, 0, -1])):
        up_vector = np.array([0, -1, 0])
    else:
        up_vector = np.cross(
            sampling_direction, np.cross(np.array([0, 0, 1]), sampling_direction)
        )
        up_vector = up_vector / np.linalg.norm(up_vector)
    if left_hand:
        up_vector = -up_vector
    up_vector = ",".join(map(str, up_vector.tolist()))
    if len(modifier_set) > 1:
        raise ValueError("Multiple modifiers")
    header = f"#@rfluxmtx h={basis} u={up_vector}\n"
    if out is not None:
        header += f'#@rfluxmtx o="{out}"\n\n'
    if source == "glow":
        source_prim = pr.Primitive("void", source, source_modifier, [], [1, 1, 1, 0])
        header += str(source_prim)
    elif source == "light":
        source_prim = pr.Primitive("void", source, source_modifier, [], [1, 1, 1])
        header += str(source_prim)
    content = ""
    for prim in surfaces:
        if prim.identifier in modifier_set:
            _identifier = "discarded"
        else:
            _identifier = prim.identifier
        _modifier = source_modifier
        if offset is not None:
            poly = parse_polygon(prim)
            offset_vec = poly.normal * offset
            moved_pts = [pt + offset_vec for pt in poly.vertices]
            _real_args = Polygon(moved_pts).vertices.flatten().tolist()
        else:
            _real_args = prim.fargs
        new_prim = pr.Primitive(
            _modifier, prim.ptype, _identifier, prim.sargs, _real_args
        )
        content += str(new_prim) + "\n"

    return header + content


def surfaces_view_factor(
    surfaces: list[pr.Primitive],
    env: list[pr.Primitive],
    ray_count: int = 10000,
) -> dict[str, dict[str, list[float]]]:
    """Calculate surface-to-surface view factors using rfluxmtx.

    Computes geometric view factors between surfaces, representing the fraction
    of radiation leaving one surface that directly reaches another surface.

    Args:
        surfaces: List of surface primitives to calculate view factors for.
            Surface normals must face outward (away from the surface).
        env: List of environment surface primitives that surfaces are exposed to.
            Surface normals must face inward (toward the calculation domain).
        ray_count: Number of rays to spawn from each surface for Monte Carlo integration.
            Higher values improve accuracy but increase computation time.

    Returns:
        Nested dictionary where outer keys are source surface identifiers and
        inner keys are target surface identifiers, with values being RGB view factors.
    """
    view_factors = {}
    surfaces_env = env + surfaces
    for idx, surface in enumerate(surfaces):
        sender = SurfaceSender([surface], basis="u")
        rest_of_surfaces = surfaces[:idx] + surfaces[idx + 1 :]
        rest_of_surface_polygons = [parse_polygon(s) for s in rest_of_surfaces]
        rest_of_surfaces_flipped = [
            polygon_primitive(p.flip(), s.modifier, s.identifier)
            for s, p in zip(rest_of_surfaces, rest_of_surface_polygons)
        ]
        receivers = [
            SurfaceReceiver([s], basis="u")
            for s in env + rest_of_surfaces_flipped
            if s.ptype in ("polygon", "ring")
        ]
        mat = Matrix(sender, receivers, octree=None, surfaces=surfaces_env)
        mat.generate(params=["-c", f"{ray_count}"])
        view_factors[surface.identifier] = {
            r.surfaces[0].identifier: [i][0] for r, i in zip(mat.receivers, mat.array)
        }
    return view_factors



import copy

import epmodel
import epmodel.epmodel as epm

from frads.window import GlazingSystem


class EnergyPlusModel(epmodel.EnergyPlusModel):
    """EnergyPlus Model object

    Attributes:
        walls_window: list of walls with windows
        floors: list of floors
        lighting_zone: list of lighting zones
        zones: list of zones
    """

    @property
    def window_walls(self) -> list[str]:
        """Get list of walls with windows."""
        if self.fenestration_surface_detailed is None:
            return []
        wndo_walls = {
            srf.building_surface_name
            for srf in self.fenestration_surface_detailed.values()
        }
        return list(wndo_walls)

    @property
    def floors(self):
        """Get all of the floor names."""
        floors = []
        if self.building_surface_detailed is None:
            return []
        for k, v in self.building_surface_detailed.items():
            if v.surface_type == epm.SurfaceType.floor:
                floors.append(k)
        return floors

    def add_glazing_system(self, glzsys: GlazingSystem):
        """Add glazing system to EnergyPlusModel's epjs dictionary.

        Args:
            glzsys: GlazingSystem object

        Raises:
            ValueError: If solar and photopic results are not computed.

        Examples:
            >>> model = load_energyplus_model(Path("model.idf"))
            >>> model.add_glazing_system(glazing_system1)
        """

        name = glzsys.name
        gap_inputs: list[epm.WindowMaterialGas | epm.WindowMaterialGasMixture] = []
        for i, gap in enumerate(glzsys.gaps):
            if len(gap.gas) == 1:
                gap_inputs.append(
                    epm.WindowMaterialGas(
                        gas_type=gap.gas[0].gas.capitalize(),
                        thickness=gap.thickness_m,
                    )
                )
            else:
                gas_dict = {
                    "thickness": gap.thickness_m,
                    "number_of_gases_in_mixture": len(gap.gas),
                }
                for idx, gas in enumerate(gap.gas):
                    gas_dict[f"gas_{idx + 1}_type"] = gas.gas.capitalize()
                    gas_dict[f"gas_{idx + 1}_fraction"] = gas.ratio
                gap_inputs.append(epm.WindowMaterialGasMixture.model_validate(gas_dict))

        layer_inputs: list[epmodel.ConstructionComplexFenestrationStateLayerInput] = []
        for i, layer in enumerate(glzsys.layers):
            layer_inputs.append(
                epmodel.ConstructionComplexFenestrationStateLayerInput(
                    name=f"{glzsys.name}_layer_{i}",
                    product_type=layer.product_type,
                    thickness=layer.thickness_m,
                    conductivity=layer.conductivity,
                    emissivity_front=layer.emissivity_front,
                    emissivity_back=layer.emissivity_back,
                    infrared_transmittance=layer.ir_transmittance,
                    directional_absorptance_front=glzsys.solar_front_absorptance[i],
                    directional_absorptance_back=glzsys.solar_back_absorptance[i],
                    top_opening_multiplier=layer.opening_multipliers.top,
                    bottom_opening_multiplier=layer.opening_multipliers.bottom,
                    left_side_opening_multiplier=layer.opening_multipliers.left,
                    right_side_opening_multiplier=layer.opening_multipliers.right,
                    front_opening_multiplier=layer.opening_multipliers.front,
                    slat_width=layer.slat_width_m,
                    slat_spacing=layer.slat_spacing_m,
                    slat_thickness=layer.slat_thickness_m,
                    slat_angle=layer.slat_angle_deg,
                    slat_conductivity=layer.slat_conductivity,
                    slat_curve=layer.slat_curve_m,
                )
            )
        input = epmodel.ConstructionComplexFenestrationStateInput(
            gaps=gap_inputs,
            layers=layer_inputs,
            solar_reflectance_back=glzsys.solar_back_reflectance,
            solar_transmittance_front=glzsys.solar_front_transmittance,
            visible_reflectance_back=glzsys.visible_back_reflectance,  # This is correct, there is typo in schema
            visible_transmittance_front=glzsys.visible_front_transmittance,
        )
        self.add_construction_complex_fenestration_state(name, input)

    def add_lighting(self, zone: str, lighting_level: float, replace: bool = False):
        """Add lighting object to EnergyPlusModel's epjs dictionary.

        Args:
            zone: Zone name to add lighting to.
            lighting_level: Lighting level in Watts.
            replace: If True, replace existing lighting object in zone.

        Raises:
            ValueError: If zone not found in model.
            ValueError: If lighting already exists in zone and replace is False.

        Examples:
            >>> model.add_lighting("Zone1", 10)
        """
        if self.zone is None:
            raise ValueError("Zone not found in model.")
        if zone not in self.zone:
            raise ValueError(f"{zone} not found in model.")
        if self.lights is None:
            raise ValueError("Lights not found in model.")
        dict2 = copy.deepcopy(self.lights)

        if self.lights is not None:
            for k, v in dict2.items():
                if v.zone_or_zonelist_or_space_or_spacelist_name == zone:
                    if replace:
                        del self.lights[k]
                    else:
                        raise ValueError(
                            f"Lighting already exists in zone = {zone}. "
                            "To replace, set replace=True."
                        )

        # Add lighting schedule type limit to epjs dictionary
        self.add(
            "schedule_type_limits",
            "on_off",
            epm.ScheduleTypeLimits(
                lower_limit_value=0,
                upper_limit_value=1,
                numeric_type=epm.NumericType.discrete,
                unit_type=epm.UnitType.availability,
            ),
        )

        # Add lighting schedule to epjs dictionary
        self.add(
            "schedule_constant",
            "constant_on",
            epm.ScheduleConstant(
                schedule_type_limits_name="on_off",
                hourly_value=1,
            ),
        )

        # Add lighting to epjs dictionary
        self.add(
            "lights",
            zone,
            epm.Lights(
                design_level_calculation_method=epm.DesignLevelCalculationMethod.lighting_level,
                fraction_radiant=0,
                fraction_replaceable=1,
                fraction_visible=1,
                lighting_level=lighting_level,
                return_air_fraction=0,
                schedule_name="constant_on",
                zone_or_zonelist_or_space_or_spacelist_name=zone,
            ),
        )

    def add_output(
        self,
        output_type: str,
        output_name: str,
        reporting_frequency: str = "Timestep",
    ):
        """Add an output variable or meter to the epjs dictionary.

        Args:
            output_type: Type of the output. "variable" or "meter".
            output_name: Name of the output variable or meter.
            reporting_frequency: Reporting frequency of the output variable or meter.

        Raises:
            ValueError: If output_type is not "variable" or "meter".

        Examples:
            >>> model.add_output("Zone Mean Air Temperature", "variable")
            >>> model.add_output("Cooling:Electricity", "meter")
        """

        if output_type == "variable":
            self._add_output_variable(output_name, reporting_frequency)
        elif output_type == "meter":
            self._add_output_meter(output_name, reporting_frequency)
        else:
            raise ValueError("output_type must be 'variable' or 'meter'.")

    def _add_output_variable(self, output_name: str, reporting_frequency):
        """Add an output variable to the epjs dictionary.

        Args:
            output_name: Name of the output variable.
            reporting_frequency: Reporting frequency of the output variable.
        """
        i = 1
        if self.output_variable is None:
            self.output_variable = {}
        for output in self.output_variable.values():
            i += 1
            if output.variable_name == output_name:
                break
        else:
            self.output_variable[f"Output:Variable {i}"] = epm.OutputVariable(
                key_value="*",
                reporting_frequency=reporting_frequency,
                variable_name=output_name,
            )

    def _add_output_meter(self, output_name: str, reporting_frequency):
        """Add an output meter to the epjs dictionary.

        Args:
            output_name: Name of the output meter.
            reporting_frequency: Reporting frequency of the output meter.
        """
        i = 1
        if self.output_meter is None:
            self.output_meter = {}
        for output in self.output_meter.values():
            i += 1
            if output.key_name == output_name:
                break
        else:
            self.output_meter[f"Output:Meter {i}"] = epm.OutputMeter(
                key_name=output_name,
                reporting_frequency=reporting_frequency,
            )



"""Convert an EnergyPlus epJSON file into Radiance model[s]."""

from dataclasses import dataclass
import logging
import math

import pyradiance as pr
import numpy as np
from frads.geom import (
    angle_between,
    Polygon,
    polygon_center,
    polygon_primitive,
    gen_grid,
)
from frads.eplus_model import EnergyPlusModel
from epmodel.epmodel import (
    BuildingSurfaceDetailed,
    FenestrationSurfaceDetailed,
    Material,
    MaterialNoMass,
    ShadingZoneDetailed,
    SurfaceType,
    WindowMaterialGlazing,
    WindowMaterialSimpleGlazingSystem,
)


logger: logging.Logger = logging.getLogger("frads.epjson2rad")


OMEGAS = {
    "kf": [
        0.0238639257641843,
        0.02332285973188108,
        0.02332285973188108,
        0.02332285973188108,
        0.02332285973188108,
        0.02332285973188108,
        0.02332285973188108,
        0.02332285973188108,
        0.02332285973188108,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.021916319185673504,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.023622221568953158,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.022385166034409335,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.02238516603440936,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.019685184640794308,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.021916319185673487,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
        0.0175372336349361,
    ],
    "kh": [
        0.04025940809134382,
        0.038724862132931824,
        0.038724862132931824,
        0.038724862132931824,
        0.038724862132931824,
        0.038724862132931824,
        0.038724862132931824,
        0.038724862132931824,
        0.038724862132931824,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04640756737178026,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.04662852131277809,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.03866539339025036,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04533939830955456,
        0.04280163786238007,
        0.04280163786238007,
        0.04280163786238007,
        0.04280163786238007,
    ],
    "kq": [
        0.07688024442411853,
        0.07132814589069367,
        0.07132814589069367,
        0.07132814589069367,
        0.07132814589069367,
        0.07132814589069367,
        0.07132814589069367,
        0.07132814589069367,
        0.07132814589069367,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08150924303937566,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.08302065811560476,
        0.06496605352254502,
        0.06496605352254502,
        0.06496605352254502,
        0.06496605352254502,
        0.06496605352254502,
        0.06496605352254502,
        0.06496605352254502,
        0.06496605352254502,
    ],
}


@dataclass
class SurfaceWithNamedFenestrations:
    """A surface with fenestrations."""

    surface: BuildingSurfaceDetailed
    fenestrations: dict[str, FenestrationSurfaceDetailed]


@dataclass
class EPlusOpaqueMaterial:
    """EnergyPlus Opaque material data container."""

    name: str
    roughness: str
    solar_absorptance: float
    visible_absorptance: float
    visible_reflectance: float
    primitive: pr.Primitive
    thickness: float = 0.0


@dataclass
class EPlusWindowMaterial:
    """EnergyPlus regular window material data container."""

    name: str
    visible_transmittance: float
    primitive: pr.Primitive
    thickness: float = 0.0


@dataclass
class EPlusConstruction:
    """EnergyPlus construction data container."""

    name: str
    type: str
    layers: list
    thickness: float


@dataclass
class EPlusOpaqueSurface:
    """EnergyPlus opaque surface data container."""

    name: str
    type: str
    polygon: Polygon
    construction: str
    boundary: str
    sun_exposed: bool
    zone: str
    fenestrations: list


def get_dict_only_value(d: None | dict):
    """Get the only value in a dictionary."""
    if d is None:
        raise ValueError("Object is None.")
    if len(d) != 1:
        raise ValueError("More than one value in the dictionary.")
    return next(iter(d.values()))


def tmit2tmis(tmit: float) -> float:
    """Convert from transmittance to transmissivity."""
    a = 0.0072522239
    b = 0.8402528435
    c = 0.9166530661
    d = 0.0036261119
    tmis = ((math.sqrt(a * tmit**2 + b) - c) / d) / tmit
    return max(0, min(tmis, 1))


def fenestration_to_polygon(fen: FenestrationSurfaceDetailed) -> Polygon:
    """Convert a fenestration_surface_detailed surface to a polygon."""
    vertices = [
        np.array(
            (
                fen.vertex_1_x_coordinate,
                fen.vertex_1_y_coordinate,
                fen.vertex_1_z_coordinate,
            )
        ),
        np.array(
            (
                fen.vertex_2_x_coordinate,
                fen.vertex_2_y_coordinate,
                fen.vertex_2_z_coordinate,
            )
        ),
        np.array(
            (
                fen.vertex_3_x_coordinate,
                fen.vertex_3_y_coordinate,
                fen.vertex_3_z_coordinate,
            )
        ),
    ]
    if fen.vertex_4_x_coordinate is not None:
        vertices.append(
            np.array(
                (
                    fen.vertex_4_x_coordinate,
                    fen.vertex_4_y_coordinate,
                    fen.vertex_4_z_coordinate,
                )
            )
        )
    return Polygon(vertices)


def surface_to_polygon(srf: BuildingSurfaceDetailed) -> Polygon:
    """Convert a building_surface_detailed surface to a polygon."""
    if srf.vertices is None:
        raise ValueError("Surface has no vertices.")
    vertices = [
        np.array(
            (
                v.vertex_x_coordinate,
                v.vertex_y_coordinate,
                v.vertex_z_coordinate,
            )
        )
        for v in srf.vertices
    ]
    return Polygon(vertices)


def exterior_shading_to_polygon(srf: ShadingZoneDetailed) -> Polygon:
    vertex = [
        np.array((v.vertex_x_coordinate, v.vertex_y_coordinate, v.vertex_z_coordinate))
        for v in srf.vertices
    ]
    return Polygon(vertex)


def thicken(
    surface: Polygon,
    windows: list[Polygon],
    thickness: float,
) -> list[Polygon]:
    """Thicken window-wall."""
    direction = surface.normal * thickness
    facade = surface.extrude(direction)[:2]
    for window in windows:
        facade.extend(window.extrude(direction)[2:])
    uniq = facade.copy()
    for idx, val in enumerate(facade):
        for rep in facade[:idx] + facade[idx + 1 :]:
            if set(map(tuple, val.vertices)) == set(map(tuple, rep.vertices)):
                uniq.remove(rep)
    return uniq


def get_construction_thickness(
    construction: EPlusConstruction, materials: dict
) -> float:
    """Get construction total thickness."""
    layer_thickness = []
    for layer in construction.layers:
        layer_thickness.append(materials[layer.lower()].thickness)
    return sum(layer_thickness)


def check_outward(polygon: Polygon, zone_center: np.ndarray) -> bool:
    """Check whether a surface is facing outside."""
    outward = True
    angle2center = angle_between(polygon.normal, zone_center - polygon.centroid)
    if angle2center < math.pi / 4:
        outward = False
    return outward


def parse_material(name: str, material: Material) -> EPlusOpaqueMaterial:
    """Parser EP Material."""
    name = name.replace(" ", "_")
    roughness = material.roughness.value
    thickness = material.thickness
    solar_absorptance = material.solar_absorptance or 0.7
    visible_absorptance = material.visible_absorptance or 0.7
    visible_reflectance = round(1 - visible_absorptance, 2)
    primitive = pr.Primitive(
        "void",
        "plastic",
        name,
        [],
        [visible_reflectance, visible_reflectance, visible_reflectance, 0, 0],
    )
    return EPlusOpaqueMaterial(
        name,
        roughness,
        solar_absorptance,
        visible_absorptance,
        visible_reflectance,
        primitive,
        thickness,
    )


def parse_material_no_mass(name: str, material: MaterialNoMass) -> EPlusOpaqueMaterial:
    """Parse EP Material:NoMass"""
    name = name.replace(" ", "_")
    roughness = material.roughness.value
    solar_absorptance = material.solar_absorptance or 0.7
    visible_absorptance = material.visible_absorptance or 0.7
    visible_reflectance = round(1 - visible_absorptance, 2)
    primitive = pr.Primitive(
        "void",
        "plastic",
        name,
        [],
        [visible_reflectance, visible_reflectance, visible_reflectance, 0, 0],
    )
    return EPlusOpaqueMaterial(
        name,
        roughness,
        solar_absorptance,
        visible_absorptance,
        visible_reflectance,
        primitive,
    )


def parse_window_material_simple_glazing_system(
    name: str, material: WindowMaterialSimpleGlazingSystem
) -> EPlusWindowMaterial:
    """Parse EP WindowMaterial:SimpleGlazingSystem"""
    identifier = name.replace(" ", "_")
    shgc = material.solar_heat_gain_coefficient
    tmit = material.visible_transmittance or shgc
    tmis = tmit2tmis(tmit)
    primitive = pr.Primitive("void", "glass", identifier, [], [tmis, tmis, tmis])
    return EPlusWindowMaterial(identifier, tmit, primitive)


def parse_window_material_glazing(
    name: str, material: WindowMaterialGlazing
) -> EPlusWindowMaterial:
    """Parse EP WindowMaterial:Glazing"""
    default_tmit = 0.6
    identifier = name.replace(" ", "_")
    if material.optical_data_type.value.lower() == "bsdf":
        tmit = 1.0
    else:
        tmit = material.visible_transmittance_at_normal_incidence or default_tmit
    tmis = tmit2tmis(tmit)
    primitive = pr.Primitive("void", "glass", identifier, [], [tmis, tmis, tmis])
    return EPlusWindowMaterial(identifier, tmit, primitive)


def parse_construction_complex_fenestration_state(
    epmodel: EnergyPlusModel,
) -> tuple:
    """Parser EP Construction:ComplexFenestrationState."""
    construction = epmodel.construction_complex_fenestration_state
    if construction is None:
        return {}, {}
    cfs = {}
    matrices = {}
    for key, val in construction.items():
        names = {
            "tvf": val.visible_optical_complex_front_transmittance_matrix_name,
            "rvb": val.visible_optical_complex_back_transmittance_matrix_name,
            "tsf": val.solar_optical_complex_front_transmittance_matrix_name,
            "rsb": val.solar_optical_complex_back_reflectance_matrix_name,
        }
        bsdf = {
            key: [v.value for v in epmodel.matrix_two_dimension[name].values]
            for key, name in names.items()
        }
        ncs = {
            key: epmodel.matrix_two_dimension[name].number_of_columns
            for key, name in names.items()
        }
        nrs = {
            key: epmodel.matrix_two_dimension[name].number_of_rows
            for key, name in names.items()
        }
        matrices[key] = {
            key: {"ncolumns": ncs[key], "nrows": nrs[key], "values": bsdf[key]}
            for key in names
        }
        cfs[key] = EPlusConstruction(key, "cfs", [], 0)
    return cfs, matrices


class EnergyPlusToRadianceModelConverter:
    """Convert EnergyPlus model to Radiance model."""

    def __init__(self, ep_model: EnergyPlusModel):
        self.model = ep_model
        self._validate_ep_model()
        self.constructions: dict = {}
        self.materials: dict = {}
        self.matrices: dict = {}

    def parse(self) -> dict[str, dict]:
        """Parse EnergyPlus model."""
        self.materials = self._parse_materials()
        self.constructions = self._parse_construction()

        zones = {zname: self._process_zone(zname) for zname in self.model.zone}

        return zones

    def _parse_construction(self) -> dict:
        """Parse EnergyPlus construction"""
        if self.model.construction is None:
            raise ValueError("No construction found in the model.")
        constructions = {}
        for cname, clayer in self.model.construction.items():
            layers = list(clayer.model_dump(exclude_none=True).values())
            constructions[cname] = EPlusConstruction(
                cname,
                "default",
                layers,
                sum(self.materials[layer.lower()].thickness for layer in layers),
            )
        cfs, matrices = parse_construction_complex_fenestration_state(self.model)
        for key, val in matrices.items():
            nested = []
            mtx = val["tvf"]
            for i in range(0, len(mtx["values"]), mtx["nrows"]):
                nested.append(mtx["values"][i : i + mtx["ncolumns"]])
            # Convert from BSDF to transmission matrix
            if mtx["ncolumns"] == 145:
                solid_angles = OMEGAS["kf"]
            elif mtx["ncolumns"] == 73:
                solid_angles = OMEGAS["kh"]
            elif mtx["ncolumns"] == 41:
                solid_angles = OMEGAS["kq"]
            else:
                raise KeyError("Unknown bsdf basis")
            assert len(solid_angles) == len(nested)
            assert len(solid_angles) == len(nested[0])
            self.matrices[key] = {
                "matrix_data": [
                    [
                        [ele * omg, ele * omg, ele * omg]
                        for ele, omg in zip(row, solid_angles)
                    ]
                    for row in nested
                ]
            }
        constructions.update(cfs)
        return constructions

    def _parse_materials(self) -> dict:
        """Parse EnergyPlus materials."""
        materials = {}
        material_keys = [
            "material",
            "material_no_mass",
            "window_material_simple_glazing_system",
            "window_material_glazing",
        ]
        for key in material_keys:
            func = f"parse_{key}".lower()
            if (mdict := getattr(self.model, key)) is not None:
                for name, material in mdict.items():
                    materials[name.lower()] = globals()[func](name, material)
        return materials

    def _process_zone(self, zone_name: str) -> dict:
        """Process a zone given the zone name.

        Args:
            zone_name: Zone name.

        Returns:
            dict: A dictionary of scene (static surfaces) and windows.
        """
        scene: list[bytes] = []
        windows: dict[str, dict[str, bytes]] = {}
        sensors: dict[str, dict] = {}
        surfaces = self._collect_zone_surfaces(zone_name)
        fenestrations = self._collect_zone_fenestrations(surfaces)
        if fenestrations == {}:  # no fenestrations in the zone
            return {}
        surfaces_fenestrations = self._pair_surfaces_fenestrations(
            surfaces, fenestrations
        )
        surface_polygons = [surface_to_polygon(srf) for srf in surfaces.values()]
        center = polygon_center(*surface_polygons)
        view_direction = np.array([0.0, 0.0, 0.0])
        for sname, swnf in surfaces_fenestrations.items():
            opaque_surface_polygon = surface_to_polygon(swnf.surface)
            (
                _surface,
                _surface_fenestrations,
                window_polygons,
            ) = self._process_surface(
                sname,
                opaque_surface_polygon,
                swnf.surface.construction_name,
                swnf.fenestrations,
                center,
            )
            scene.extend(_surface)
            windows.update(_surface_fenestrations)
            if swnf.surface.surface_type == SurfaceType.floor:
                sensors[sname] = {
                    "data": gen_grid(
                        polygon=opaque_surface_polygon,
                        height=0.76,
                        spacing=0.61,
                    )
                }
            for window_polygon in window_polygons:
                view_direction += window_polygon.area
        view_direction *= -1
        view = pr.create_default_view()
        view.vp = tuple(center)
        view.vdir = tuple(view_direction)
        view.type = "a"
        view.horiz = 180
        view.vert = 180
        sensors[zone_name] = {"data": [center.tolist() + view_direction.tolist()]}

        return {
            "scene": {"bytes": b" ".join(scene)},
            "windows": windows,
            "materials": {
                "bytes": b" ".join(
                    mat.primitive.bytes for mat in self.materials.values()
                ),
                "matrices": self.matrices,
            },
            "sensors": sensors,
            "views": {zone_name: {"view": view}},
        }

    def _process_surface(
        self,
        surface_name: str,
        surface_polygon: Polygon,
        surface_construction_name: str,
        fenestrations: dict[str, FenestrationSurfaceDetailed],
        zone_center: np.ndarray,
    ) -> tuple[list[bytes], dict[str, dict[str, bytes]], list[Polygon]]:
        """Process a surface in a zone.

        Args:
            surface_name: Surface name.
            surface_polygon: Surface polygon.
            surface_construction_name: Surface construction name.
            fenestrations: Fenestrations in the surface.
            zone_center: Zone center.

        Returns:
            tuple[list[bytes], dict[str, dict[str, bytes]], list[Polygon]]: A tuple of scene, windows, and window polygons.
        """
        scene: list[bytes] = []
        windows: dict[str, dict] = {}
        window_polygons: list[Polygon] = []
        opaque_surface_name = surface_name.replace(" ", "_")
        if not check_outward(surface_polygon, zone_center):
            surface_polygon = surface_polygon.flip()
        for fname, fene in fenestrations.items():
            fenestration_polygon, window = self._process_fenestration(
                fname, fene, zone_center
            )
            window_polygons.append(fenestration_polygon)
            surface_polygon -= fenestration_polygon
            windows[fname] = {"bytes": window.bytes, "polygon": fenestration_polygon}
            if fene.construction_name in self.matrices:
                windows[fname]["matrix_name"] = fene.construction_name
        # polygon to primitive
        construction = self.constructions[surface_construction_name]
        inner_material_name = construction.layers[-1].replace(" ", "_")
        scene.append(
            polygon_primitive(
                surface_polygon, inner_material_name, opaque_surface_name
            ).bytes
        )

        # process exterior shading
        shading_surfaces: dict = self._collect_shading_surfaces(surface_name)
        outer_material_name = construction.layers[0].replace(" ", "_")
        for sname, shade in shading_surfaces.items():
            shading_polygon = exterior_shading_to_polygon(shade)
            scene.append(
                polygon_primitive(shading_polygon, outer_material_name, sname).bytes
            )

        # extrude the surface by thickness
        if fenestrations != {}:
            facade = thicken(surface_polygon, window_polygons, construction.thickness)
            scene.append(
                polygon_primitive(
                    facade[1], outer_material_name, f"ext_{opaque_surface_name}"
                ).bytes
            )
            for idx in range(2, len(facade)):
                scene.append(
                    polygon_primitive(
                        facade[idx],
                        inner_material_name,
                        f"sill_{opaque_surface_name}.{idx}",
                    ).bytes
                )

        return scene, windows, window_polygons

    def _collect_zone_surfaces(
        self, zone_name: str
    ) -> dict[str, BuildingSurfaceDetailed]:
        if self.model.building_surface_detailed is None:
            return {}
        return {
            sname: srf
            for sname, srf in self.model.building_surface_detailed.items()
            if srf.zone_name == zone_name
        }

    def _collect_zone_fenestrations(
        self, zone_surfaces: dict
    ) -> dict[str, FenestrationSurfaceDetailed]:
        if self.model.fenestration_surface_detailed is None:
            return {}
        return {
            fname: fen
            for fname, fen in self.model.fenestration_surface_detailed.items()
            if fen.building_surface_name in zone_surfaces
        }

    def _collect_shading_surfaces(self, surface_name: str):
        if self.model.shading_zone_detailed is None:
            return {}
        return {
            sname: shade
            for sname, shade in self.model.shading_zone_detailed.items()
            if shade.base_surface_name == surface_name
        }

    def _pair_surfaces_fenestrations(
        self,
        zone_surfaces: dict[str, BuildingSurfaceDetailed],
        zone_fenestrations: dict[str, FenestrationSurfaceDetailed],
    ) -> dict[str, SurfaceWithNamedFenestrations]:
        surface_fenestrations = {}
        for sname, srf in zone_surfaces.items():
            named_fen = {}
            for fname, fen in zone_fenestrations.items():
                if fen.building_surface_name == sname:
                    named_fen[fname] = fen
            surface_fenestrations[sname] = SurfaceWithNamedFenestrations(srf, named_fen)
        return surface_fenestrations

    def _process_fenestration(
        self,
        name: str,
        fenestration: FenestrationSurfaceDetailed,
        zone_center: np.ndarray,
    ) -> tuple[Polygon, pr.Primitive]:
        fenenstration_polygon = fenestration_to_polygon(fenestration)
        if check_outward(fenenstration_polygon, zone_center):
            fenenstration_polygon = fenenstration_polygon.flip()
        _construction = self.constructions[fenestration.construction_name]
        if _construction.type == "cfs":
            window_material = "void"
        else:
            window_material = _construction.layers[0].replace(" ", "_")
        fenestration_primitive = polygon_primitive(
            fenenstration_polygon, window_material, name.replace(" ", "_")
        )
        return fenenstration_polygon, fenestration_primitive

    def _validate_ep_model(self) -> bool:
        """Validate EnergyPlus model.
        Make sure the model has at least one zone, one building surface,
        and one fenestration.

        Returns:
            bool: True if the model is valid.
        """
        valid = True
        if self.model.zone is None:
            logger.warning("No zone found in the model.")
            valid = False
        if self.model.building_surface_detailed is None:
            logger.warning("No building surface found in the model.")
            valid = False
        if self.model.fenestration_surface_detailed is None:
            logger.warning("No fenestration found in the model.")
            valid = False
        return valid


def create_settings(
    ep_model: EnergyPlusModel, epw_file: None | str, output_directory: str
) -> dict:
    """Create settings dictionary for Radiance model.

    Args:
        ep_model: EnergyPlus model.
        epw_file: EnergyPlus weather file path.
        output_directory: Output directory for Radiance files.

    Returns:
        dict: Radiance settings.
    """
    settings = {"method": "3", "sky_basis": "r1", "output_directory": output_directory}
    if epw_file:
        settings["epw_file"] = epw_file
    else:
        site = get_dict_only_value(ep_model.site_location)
        settings.update(
            {
                "latitude": site.latitude,
                "longitude": site.longitude * -1,
                "time_zone": site.time_zone * -15,
                "site_elevation": site.elevation,
            }
        )
    return settings


def epmodel_to_radmodel(
    ep_model: EnergyPlusModel,
    epw_file: None | str = None,
    add_views: bool = True,
    views: None | dict[str, dict[str, pr.View]] = None,
    output_directory: str = None,
) -> dict:
    """Convert EnergyPlus model to Radiance models where each zone is a separate model.

    Args:
        ep_model: EnergyPlus model.
        epw_file: EnergyPlus weather file path. Defaults to None.
        add_views: Add views to the model. Such views will be positioned
            at the center of the zone facing windows weighted by window
            area. Defaults to True.

    Returns:
        dict: Radiance models.

    Examples:
        >>> radmodels = epmodel_to_radmodel(ep_model, epw_file)
    """
    model_parser = EnergyPlusToRadianceModelConverter(ep_model)
    zone_models = model_parser.parse()
    settings = create_settings(ep_model, epw_file, output_directory)
    rad_models = {
        zname: {"settings": settings, "model": zmodel}
        for zname, zmodel in zone_models.items()
        if zmodel != {}
    }
    if not add_views:
        for model in rad_models.values():
            model["views"] = {}
    for model in rad_models.values():
        model["surfaces"] = {}

    return rad_models
